#######################################################################
# valencia1d formulation
#
# This is the ordinary Valencia formuatlion as first proposed in
# Banyuls+1997, Thierfelder+2011, but written in spherical coordinates
# and evaluated at θ=π/2, ϕ=0.
# In this formulation the conservative variables also contain a
# factor r which comes from the sqrt of det(induced metric).
# Furthermore, the source terms contain a 1/r term.
# These are now two things that need to be regularized when the origin
# is enclosed.
# This implementation supports non-zero shift vector.
#######################################################################


@with_signature function unscale_conservatives_valencia1d(equation::Equation)
  @accepts rD, rSr, rτ, sqrt_detγ
  D  = rD/sqrt_detγ
  Sr = rSr/sqrt_detγ
  τ  = rτ/sqrt_detγ
  @returns D, Sr, τ
end


@with_signature function flux_source_valencia1d(eq::Equation)
  @accepts D, Sr, τ, ϵ, ρ, p, vr
  @accepts α, ∂αr, βru, ∂βrrdu, sqrt_detγ,
           γrr, γθθ, γϕϕ, ∂γrrr, ∂γrθθ, ∂γrϕϕ, ∂γθϕϕ,
           Γrrr, Γrθθ, Γrϕϕ, Γθrθ, Γθϕϕ, Γϕθϕ, Γϕrϕ,
           Krr, Kθθ, Kϕϕ

  γuurr = 1/γrr
  γuuθθ = 1/γθθ
  γuuϕϕ = 1/γϕϕ
  Γkrk = Γrrr + Γθrθ + Γϕrϕ

  gtt = -α^2 + γrr*βru*βru
  gtr = γrr*βru
  grr = γrr
  guutt = -1/α^2
  guutr = βru/α^2
  guurr = γuurr - βru*βru/α^2
  guuθθ = γuuθθ
  guuϕϕ = γuuϕϕ

  vru = γuurr*vr
  v2 = vru*vr
  W = 1/sqrt(1-v2)
  ut = W/α
  ur = W*(vru - βru/α)

  h = 1 + ϵ + p/ρ
  Tuutt = ρ*h*ut*ut + p*guutt
  Tuutr = ρ*h*ut*ur + p*guutr
  Tuurr = ρ*h*ur*ur + p*guurr
  Tuuθθ = p*guuθθ
  Tuuϕϕ = p*guuϕϕ
  Tudtr = Tuutt*gtr + Tuutr*grr

  sqrt_detg = α*sqrt_detγ

  # computing flux and source terms following Thierfelder, Bernuzzi, Bruegmann, 2011

  flr_rD  = sqrt_detg*D*(vru - βru/α)
  flr_rSr = sqrt_detg*( Sr*(vru - βru/α) + p )
  flr_rτ  = sqrt_detg*( τ*(vru - βru/α) + p*vru )

  # src_D = 0

  # Tᵗᵗ (1/2 βⁱβʲ ∂ₖγᵢⱼ - α∂ₖα)
  tmp1 = Tuutt*(1/2*βru*βru*∂γrrr - α*∂αr)
  # Tᵗⁱ βʲ ∂ₖ γᵢⱼ
  tmp2 = Tudtr*βru*∂γrrr
  # Tᵗᵢ ∂ₖ βⁱ
  tmp3 = Tudtr*∂βrrdu
  # 1/2 Tⁱʲ ∂ₖ γᵢⱼ
  tmp4 = 1/2 * (Tuurr*∂γrrr + Tuuθθ*∂γrθθ + Tuuϕϕ*∂γrϕϕ)
  src_rSr = sqrt_detg*(tmp1 + tmp2 + tmp3 + tmp4)

  # Tᵗᵗ(βⁱβʲ Kᵢⱼ - βⁱ∂ᵢα)
  tmp1 = Tuutt * (βru*βru*Krr - βru*∂αr)
  # Tᵗⁱ(2βʲKᵢⱼ - ∂ᵢα)
  tmp2 = Tuutr * (2*βru*Krr - ∂αr)
  # TⁱʲKᵢⱼ
  tmp3 = Tuurr*Krr + Tuuθθ*Kθθ + Tuuϕϕ*Kϕϕ
  src_rτ = sqrt_detg*(tmp1 + tmp2 + tmp3)

  @returns #=src_rD,=# src_rSr, src_rτ, flr_rD, flr_rSr, flr_rτ
end


@with_signature function llf_valencia1d(equation::Equation)
  @accepts        rD, flr_rD, rSr, flr_rSr, rτ, flr_rτ, max_v, p
  @accepts        α, βru, γrr, sqrt_detγ
  @accepts [bdry] bdry_D, bdry_Sr, bdry_τ, bdry_max_v, bdry_ρ, bdry_vr, bdry_ϵ, bdry_p
  @accepts [bdry] nx

  vmax     = absmax(max_v, bdry_max_v)
  γuurr    = 1/γrr
  bdry_vru = γuurr * bdry_vr

  sqrt_detg = α*sqrt_detγ

  bdry_rD  = sqrt_detγ*bdry_D
  bdry_rSr = sqrt_detγ*bdry_Sr
  bdry_rτ  = sqrt_detγ*bdry_τ

  bdry_flr_rD  = sqrt_detg * bdry_D * (bdry_vru - βru/α)
  bdry_flr_rSr = sqrt_detg * ( bdry_Sr * (bdry_vru - βru/α) + bdry_p )
  bdry_flr_rτ  = sqrt_detg * ( bdry_τ * (bdry_vru - βru/α) + bdry_p * bdry_vru )

  nflr      = nx*flr_rD
  bdry_nflr = nx*bdry_flr_rD
  nflr_rD   = LLF(nflr, bdry_nflr, rD, bdry_rD, vmax)

  nflr      = nx*flr_rSr
  bdry_nflr = nx*bdry_flr_rSr
  nflr_rSr  = LLF(nflr, bdry_nflr, rSr, bdry_rSr, vmax)

  nflr      = nx*flr_rτ
  bdry_nflr = nx*bdry_flr_rτ
  nflr_rτ   = LLF(nflr, bdry_nflr, rτ, bdry_rτ, vmax)

  @returns [bdry] nflr_rD, nflr_rSr, nflr_rτ
end


@with_signature function bdryllf_valencia1d_tov(equation::Equation)
  @accepts        flr_rD, flr_rSr, flr_rτ
  @accepts [bdry] nx

  # vacuum should show no dynamics, hence, we can just copy the flux
  nflr_rD  = nx*flr_rD
  nflr_rSr = nx*flr_rSr
  nflr_rτ  = nx*flr_rτ

  @returns [bdry] nflr_rD, nflr_rSr, nflr_rτ
end
@with_signature function bdryllf_valencia1d_bondi(equation::Equation)
  @accepts        D, flr_rD, Sr, flr_rSr, τ, flr_rτ, max_v, vr, p
  @accepts        init_D, init_Sr, init_τ, init_vr, init_max_v, init_p
  @accepts        α, βru, γrr, sqrt_detγ
  @accepts [bdry] nx

  vmax     = absmax(max_v, init_max_v)
  γuurr    = 1/γrr
  init_vru = γuurr * init_vr
  vru      = γuurr * vr

  sqrt_detg = α*sqrt_detγ

  rD  = sqrt_detγ*D
  rSr = sqrt_detγ*Sr
  rτ  = sqrt_detγ*τ

  # Dirichlet conditions
  bdry_D   = -D   + 2*init_D
  bdry_Sr  = -Sr  + 2*init_Sr
  bdry_τ   = -τ   + 2*init_τ
  bdry_vru = -vru + 2*init_vru
  bdry_p   = -p   + 2*init_p
  # bdry_D   = init_D
  # bdry_Sr  = init_Sr
  # bdry_τ   = init_τ
  # bdry_vru = init_vru
  # bdry_p   = init_p

  bdry_rD  = sqrt_detγ*bdry_D
  bdry_rSr = sqrt_detγ*bdry_Sr
  bdry_rτ  = sqrt_detγ*bdry_τ

  bdry_flr_rD  = sqrt_detg * bdry_D * (bdry_vru - βru/α)
  bdry_flr_rSr = sqrt_detg * ( bdry_Sr * (bdry_vru - βru/α) + bdry_p )
  bdry_flr_rτ  = sqrt_detg * ( bdry_τ * (bdry_vru - βru/α) + bdry_p * bdry_vru )

  nflr      = nx*flr_rD
  bdry_nflr = nx*bdry_flr_rD
  nflr_rD   = LLF(nflr, bdry_nflr, rD, bdry_rD, vmax)

  nflr      = nx*flr_rSr
  bdry_nflr = nx*bdry_flr_rSr
  nflr_rSr  = LLF(nflr, bdry_nflr, rSr, bdry_rSr, vmax)

  nflr      = nx*flr_rτ
  bdry_nflr = nx*bdry_flr_rτ
  nflr_rτ   = LLF(nflr, bdry_nflr, rτ, bdry_rτ, vmax)

  @returns [bdry] nflr_rD, nflr_rSr, nflr_rτ
end


@with_signature function bdryllf_valencia1d_tov_reflected(equation::Equation)
  @accepts        D, flr_rD, Sr, flr_rSr, τ, flr_rτ, max_v, vr, p, r
  @accepts        bdry_D, bdry_Sr, bdry_τ, bdry_vr, bdry_max_v, bdry_p
  @accepts        α, βru, γrr, sqrt_detγ
  @accepts [bdry] nx

  if nx > 0

    # vacuum should show no dynamics, hence, we can just copy the flux
    nflr_rD  = nx*flr_rD
    nflr_rSr = nx*flr_rSr
    nflr_rτ  = nx*flr_rτ

  else

    # # apply cons2prim to bdry values
    # We can perform cons2prim analytically in this case, because
    # bdry_Sr = 0 implies bdry_vr = 0. Thus, W = 1.
    bdry_vr = 0
    bdry_ρ = bdry_D
    bdry_ϵ = bdry_τ/bdry_ρ
    @unpack eos = equation
    bdry_p = eos(Pressure, Density, bdry_ρ, InternalEnergy, bdry_ϵ)
    bdry_max_v = 0

    vmax     = absmax(max_v, bdry_max_v)
    γuurr    = 1/γrr
    bdry_vru = γuurr * bdry_vr
    vru      = γuurr * vr

    sqrt_detg = α*sqrt_detγ

    rD  = sqrt_detγ*D
    rSr = sqrt_detγ*Sr
    rτ  = sqrt_detγ*τ
    bdry_rD  = sqrt_detγ*bdry_D
    bdry_rSr = sqrt_detγ*bdry_Sr
    bdry_rτ  = sqrt_detγ*bdry_τ

    # # Dirichlet conditions
    # bdry_D   = -D   + 2*bdry_D
    # bdry_Sr  = -Sr  + 2*bdry_Sr
    # bdry_τ   = -τ   + 2*bdry_τ
    # bdry_vru = -vru + 2*bdry_vru
    # bdry_p   = -p   + 2*bdry_p

    bdry_flr_rD  = sqrt_detg * bdry_D * (bdry_vru - βru/α)
    bdry_flr_rSr = sqrt_detg * ( bdry_Sr * (bdry_vru - βru/α) + bdry_p )
    bdry_flr_rτ  = sqrt_detg * ( bdry_τ * (bdry_vru - βru/α) + bdry_p * bdry_vru )

    nflr      = nx*flr_rD
    bdry_nflr = nx*bdry_flr_rD
    nflr_rD   = LLF(nflr, bdry_nflr, rD, bdry_rD, vmax)

    nflr      = nx*flr_rSr
    bdry_nflr = nx*bdry_flr_rSr
    nflr_rSr  = LLF(nflr, bdry_nflr, rSr, bdry_rSr, vmax)

    nflr      = nx*flr_rτ
    bdry_nflr = nx*bdry_flr_rτ
    nflr_rτ   = LLF(nflr, bdry_nflr, rτ, bdry_rτ, vmax)

  end

  @returns [bdry] nflr_rD, nflr_rSr, nflr_rτ
end


function impose_symmetry_bdry_values_valencia1d!(mesh::Mesh1d{SpectralElement})

  @unpack cache = mesh
  @unpack D, Sr, τ = get_static_variables(mesh)
  @unpack bdry_D, bdry_Sr, bdry_τ = get_bdry_variables(mesh)

  L = layout(mesh)
  Npts, _ = L
  vD  = dg1d.vreshape(D, L)
  vSr = dg1d.vreshape(Sr, L)
  vτ  = dg1d.vreshape(τ, L)

  bdryidx = 1

  # even variables are interpolated such that derivative vanishes at r=0
  for (bdryvar,var) in ((bdry_D,vD), (bdry_τ,vτ))
    tmp = 0
    for j = 2:Npts
      tmp += mesh.element.D[1,j] * var[j,1]
    end
    bdryvar[bdryidx] = -tmp/mesh.element.D[1,1]
  end
  # odd variables are set to zero at r=0
  bdry_Sr[bdryidx] = 0

end
impose_symmetry_bdry_values_doublecartoon!(mesh) = TODO(mesh)


function impose_symmetry!(P::Project{:valencia1d}, mesh::Mesh)

  @unpack cache = mesh
  @unpack D, Sr, τ = get_static_variables(mesh)

  if singularity_on_grid(mesh)
    L = layout(mesh)
    Npts, _ = L

    vD  = dg1d.vreshape(D, L)
    vSr = dg1d.vreshape(Sr, L)
    vτ  = dg1d.vreshape(τ, L)

    if !ismirrored(mesh)
      # even variables are interpolated such that derivative vanishes at r=0
      for var in (vD, vτ)
        tmp = 0
        for j = 2:Npts
          tmp += mesh.element.D[1,j] * var[j,1]
        end
        # Jacobian cancels here, because its constant throughout
        var[1,1] = -tmp/mesh.element.D[1,1]
      end
      # odd variables are set to zero at r=0
      Sr[1] = 0
    else
      TODO()
    end
  else
    idx = find_cell_index(0.0, mesh)
    buf = zeros(Float64,dg1d.npoints(mesh))
    # symmetric
    for var in (D,τ)
      vvar = dg1d.vreshape(var, layout(mesh))
      @views @. begin
        buf = vvar[:,idx]
        vvar[:,idx] = 1/2*(buf+buf[end:-1:1])
      end
    end
    # anti-symmetric
    vvar = dg1d.vreshape(Sr, layout(mesh))
    @views @. begin
      buf = vvar[:,idx]
      vvar[:,idx] = 1/2*(buf-buf[end:-1:1])
    end
  end

  return
end


function impose_symmetry_sources!(P::Project{:valencia1d}, mesh::Mesh)

  singularity_on_grid(mesh) || return

  @unpack r = get_static_variables(mesh.cache)
  @unpack src_rD, src_rSr, src_rτ = get_static_variables(mesh.cache)
  L = layout(mesh)
  Npts, _ = L
  v_src_rD  = dg1d.vreshape(src_rD, L)
  v_src_rτ  = dg1d.vreshape(src_rτ, L)
  # even variables are interpolated such that derivative vanishes at r=0
  for var in (v_src_rD, v_src_rτ)
    tmp = 0
    for j = 2:Npts
      tmp += mesh.element.D[1,j] * var[j,1]
    end
    var[1,1] = -tmp/mesh.element.D[1,1]
  end
  # odd variables are set to zero at r=0
  src_rSr[1] = 0
  return
end


#######################################################################
#                                 LDG                                 #
#######################################################################


@with_signature function ldg_nflux_valencia1d(eq::Equation)
  @accepts        rD, rSr, rτ, sqrt_detγ, α
  @accepts [bdry] bdry_D, bdry_Sr, bdry_τ
  @accepts [bdry] nx
  sqrt_detg = α * sqrt_detγ
  bdry_rD  = sqrt_detg*bdry_D
  bdry_rSr = sqrt_detg*bdry_Sr
  bdry_rτ  = sqrt_detg*bdry_τ
  nflr_rD  = 0.5*nx*(rD+bdry_rD)
  nflr_rSr = 0.5*nx*(rSr+bdry_rSr)
  nflr_rτ  = 0.5*nx*(rτ+bdry_rτ)
  @returns [bdry] nflr_rD, nflr_rSr, nflr_rτ
end


@with_signature function ldg_bdryllf_valencia1d(eq::Equation)
  @accepts        rD, rSr, rτ
  @accepts [bdry] nx
  nflr_rD  = nx*rD
  nflr_rSr = nx*rSr
  nflr_rτ  = nx*rτ
  @returns [bdry] nflr_rD, nflr_rSr, nflr_rτ
end


@with_signature function av_flux_covariant_valencia1d(equations::Equation)
  @accepts rhs_rD, rhs_rSr, rhs_rτ, flr_rD, flr_rSr, flr_rτ, ldg_rD, ldg_rSr, ldg_rτ, smoothed_mu
  flr_rD  += -smoothed_mu * (rhs_rD  - ldg_rD)
  flr_rSr += -smoothed_mu * (rhs_rSr - ldg_rSr)
  flr_rτ  += -smoothed_mu * (rhs_rτ  - ldg_rτ)
  @returns flr_rD, flr_rSr, flr_rτ
end


@with_signature function av_nflux_covariant_valencia1d(eq::Equation)
  @accepts        rhs_rD, rhs_rSr, rhs_rτ, ldg_rD, ldg_rSr, ldg_rτ, smoothed_mu, r
  @accepts [bdry] bdry_rhs_rD, bdry_rhs_rSr, bdry_rhs_rτ, bdry_ldg_rD, bdry_ldg_rSr, bdry_ldg_rτ, bdry_smoothed_mu
  @accepts [bdry] nflr_rD, nflr_rSr, nflr_rτ, nx

  nflr       = -nx*(rhs_rD - ldg_rD)
  bdry_nflr  = -nx*(bdry_rhs_rD - bdry_ldg_rD)
  nflr_rD   += 0.5 * (smoothed_mu*nflr + bdry_smoothed_mu*bdry_nflr)

  nflr       = -nx*(rhs_rSr - ldg_rSr)
  bdry_nflr  = -nx*(bdry_rhs_rSr - bdry_ldg_rSr)
  nflr_rSr  += 0.5 * (smoothed_mu*nflr + bdry_smoothed_mu*bdry_nflr)

  nflr       = -nx*(rhs_rτ - ldg_rτ)
  bdry_nflr  = -nx*(bdry_rhs_rτ - bdry_ldg_rτ)
  nflr_rτ   += 0.5 * (smoothed_mu*nflr + bdry_smoothed_mu*bdry_nflr)

  @returns [bdry] nflr_rD, nflr_rSr, nflr_rτ
end


@with_signature function stop_atmosphere_evolution_valencia1d(eq::Equation)
  @accepts rhs_rD, rhs_rSr, rhs_rτ, c2p_reset_atm
  if c2p_reset_atm > 0
    rhs_rD = rhs_rSr = rhs_rτ = 0.0
  end
  @returns rhs_rD, rhs_rSr, rhs_rτ
end
