#######################################################################
#                              Timestep                               #
#######################################################################

timestep(env, P, hrsc) = timestep(env, P, hrsc, env.mesh)


function compute_maxspeed(mesh, equation, cache)
  @unpack max_v = get_static_variables(cache)
  broadcast_volume!(maxspeed, equation, mesh)
  vmax = maximum(abs, max_v)
  vmax_limit = 1e4
  if vmax > vmax_limit
    @warn "Limiting timestep due to maximum speed exceeding $vmax_limit"
    vmax = vmax_limit
  end
  return vmax
end


function timestep(env, P::Project, hrsc::Maybe{HRSC.AbstractHRSC}, mesh::Mesh1d)
  vmax = compute_maxspeed(mesh, P.equation, mesh.cache)
  dl = min_grid_spacing(mesh)
  dt = dl / (vmax * dtfactor(mesh))
  return dt
end
dtfactor(mesh::Mesh1d{FVElement}) = 2
dtfactor(mesh::Mesh1d{SpectralElement}) = mesh.element.N


function timestep(env, P::Project, hrsc::HRSC.AbstractArtificialViscosity, mesh::Mesh1d)
  @unpack cache    = env
  @unpack equation = P
  @unpack max_v    = get_static_variables(cache)

  vmax = maximum(abs, max_v)
  vmax_limit = 0.9999999
  if vmax > vmax_limit
    @warn "Limiting timestep due to maximum speed $vmax exceeding $vmax_limit"
    vmax = vmax_limit
  end
  smoothed_mu = get_variable(cache, :smoothed_mu)
  mumax = dg1d.absolute_maximum(view(smoothed_mu, :))

  @unpack element = mesh
  @unpack N = element
  L, = dg1d.widths(mesh)
  K, = mesh.tree.dims
  dL = L/K

  dt = 1 / (vmax * N / dL + mumax * N^3 / dL^2)
  dt_limit = 1e-7
  if dt < dt_limit
    @warn "Limiting timestep due to dt being smaller than $dt_limit"
    dt = dt_limit
  end
  return dt
end


#######################################################################
#                                 2d                                  #
#######################################################################


function timestep(env, P, hrsc::Maybe{HRSC.AbstractHRSC}, mesh::Mesh2d)
  @unpack cache    = env
  @unpack equation = P
  @unpack max_v = get_static_variables(cache)

  broadcast_volume!(cons2prim, equation, mesh)
  broadcast_volume!(maxspeed, equation, mesh)
  i_vmax, vmax = dg1d.absolute_maximum_index(max_v)
  vmax_limit = 0.9999
  if vmax > vmax_limit
    i_cell = find_cell_index(i_vmax, mesh)
    cell = mesh.tree.cells[i_cell]
    @warn "Limiting timestep due to maximum speed exceeding $vmax_limit in cell '$cell'"
    vmax = vmax_limit
  end

  @unpack element = mesh
  @unpack N, z = element

  dx, dy = dg1d.widths(mesh.boxes[1])
  for box in view(mesh.boxes, 2:length(mesh.boxes))
    ddx, ddy = dg1d.widths(box)
    ddx < dx && (dx = ddx)
    ddy < dy && (dy = ddy)
  end
  dz = z[2]-z[1]
  dl = min(dx, dy) * dz

  dt = dl / (N * vmax)

  return dt
end


#######################################################################
#                               FV rhs                                #
#######################################################################


function rhs!(env, P::Project, hrsc::Nothing, mesh::Mesh1d{FVElement})
  if P.prms.hrsc === :fv_central
    rhs_fv_central!(P, mesh)
  elseif P.prms.hrsc === :slope_limiter
    rhs_fv_muscl!(P, mesh)
  else
    TODO(P.prms.hrsc)
  end
end


function rhs_fv_central!(P::Project, mesh::Mesh1d{FVElement})

  @unpack equation                 = P
  @unpack D, S, tau                = get_dynamic_variables(mesh)
  @unpack flx_D, flx_S, flx_tau    = get_static_variables(mesh)
  @unpack rhs_D, rhs_S, rhs_tau    = get_rhs_variables(mesh)
  @unpack nflx_D, nflx_S, nflx_tau = get_bdry_variables(mesh)
  @unpack bdry_D, bdry_S, bdry_tau = get_bdry_variables(mesh)
  @unpack dt                       = get_global_variables(mesh)

  broadcast_volume!(cons2prim_kastaun, equation, mesh)
  broadcast_volume!(maxspeed, equation, mesh)
  broadcast_volume!(flux, equation, mesh)
  broadcast_bdry!(fv_bdry_flux, equation, mesh)

  fv_update_step!(rhs_D,   D,   flx_D,   bdry_D,   nflx_D,   dt[1], mesh)
  fv_update_step!(rhs_S,   S,   flx_S,   bdry_S,   nflx_S,   dt[1], mesh)
  fv_update_step!(rhs_tau, tau, flx_tau, bdry_tau, nflx_tau, dt[1], mesh)

  return
end


function rhs_fv_muscl!(P::Project, mesh::Mesh1d{FVElement})
  @unpack equation              = P
  @unpack D, S, tau             = get_dynamic_variables(mesh)
  @unpack D1, S1, tau1,
          D2, S2, tau2          = get_static_variables(mesh)
  @unpack rhs_D, rhs_S, rhs_tau = get_rhs_variables(mesh)
  @unpack dt = get_global_variables(mesh)
  dt = dt[1]

  # SSPRK3 stepping from Hesthaven 2018 book, sec 10.2 (cf Script 10.13)
  step_fv_muscl!(P,mesh,(D,S,tau))
  @. D1   = D   + dt*rhs_D
  @. S1   = S   + dt*rhs_S
  @. tau1 = tau + dt*rhs_tau
  step_fv_muscl!(P,mesh,(D1,S1,tau1))
  @. D2   = (3*D   + D1   + dt*rhs_D)/4
  @. S2   = (3*S   + S1   + dt*rhs_S)/4
  @. tau2 = (3*tau + tau1 + dt*rhs_tau)/4
  step_fv_muscl!(P,mesh,(D2,S2,tau2))
  # atm final result is published via rhs_ variables
  @. rhs_D   = (D   + 2*D2   + 2*dt*rhs_D)/3
  @. rhs_S   = (S   + 2*S2   + 2*dt*rhs_S)/3
  @. rhs_tau = (tau + 2*tau2 + 2*dt*rhs_tau)/3

  return
end


function step_fv_muscl!(P::Project, mesh::Mesh1d{FVElement}, state)

  @unpack equation                    = P
  @unpack D, S, tau                   = get_dynamic_variables(mesh)
  @unpack rhs_D, rhs_S, rhs_tau       = get_rhs_variables(mesh)
  @unpack rho, v, eps,
          bD, bS, btau,
          rhoL, vL, epsL, pL,
          rhoR, vR, epsR, pR,
          drho_p, dv_p, deps_p,
          drho_m, dv_m, deps_m,
          nflx_D_L, nflx_S_L, nflx_tau_L,
          nflx_D_R, nflx_S_R, nflx_tau_R  = get_static_variables(mesh)
  @unpack eos = equation

  sD, sS, stau = state
  bD   .= D
  bS   .= S
  btau .= tau
  D   .= sD
  S   .= sS
  tau .= stau
  broadcast_volume!(cons2prim_kastaun, P.equation, mesh)
  broadcast_volume!(maxspeed, P.equation, mesh)
  D   .= bD
  S   .= bS
  tau .= btau

  periodic = isperiodic(mesh)
  prms_slope_limiter = (; method=P.prms.slope_limiter_method,
                          tvb_M=P.prms.slope_limiter_tvb_M)
  K, = mesh.tree.dims
  L, = dg1d.widths(mesh)
  dl = L/K

  for vars in ( (rhoL,rhoR,rho,drho_m,drho_p),
                (vL,vR,v,dv_m,dv_p),
                (epsL,epsR,eps,deps_m,deps_p) )

    wL,wR,w,dw₋,dw₊ = vars

    # compute left (-) and right (+) differences
    @turbo for k in 2:K-1
      k₋, k₊ = k-1, k+1
      w₋, w₀, w₊ = w[k₋], w[k], w[k₊]
      dw₊[k], dw₋[k] = w₊-w₀, w₀-w₋
    end
    for k in (1,K)
      k₋, k₊ = periodic_index(k-1,K), periodic_index(k+1,K)
      w₋, w₀, w₊ = w[k₋], w[k], w[k₊]
      dw₊[k], dw₋[k] = w₊-w₀, w₀-w₋
    end

    # limit slope and compute left (L) and (R) interface values
    for k in 1:K
      dwL = limit_slope(dw₊[k], dw₋[k], dl, prms_slope_limiter)
      wL[k] = w[k] - dwL/2
      wR[k] = w[k] + dwL/2
    end
  end

  for k in 1:K
    pL[k] = eos(Pressure, Density, rhoL[k], InternalEnergy, epsL[k])
    pR[k] = eos(Pressure, Density, rhoR[k], InternalEnergy, epsR[k])
  end

  # numerical fluxes
  # subscript + ... cell interior value
  # subscript - ... cell exterior value
  # nx ... pointing from interior too exterior cell
  @unpack max_v = get_static_variables(mesh)
  for k in 2:K
    # on left cell interfaces
    nx            = -1
    k₊, k₋        = k, k-1
    v₊, v₋        = vL[k₊], vR[k₋]
    rho₊, rho₋    = rhoL[k₊], rhoR[k₋]
    eps₊, eps₋    = epsL[k₊], epsR[k₋]
    p₊, p₋        = pL[k₊], pR[k₋]
    D₊, D₋        = Formulae.D(rho₊,v₊,eps₊,p₊), Formulae.D(rho₋,v₋,eps₋,p₋)
    S₊, S₋        = Formulae.S(rho₊,v₊,eps₊,p₊), Formulae.S(rho₋,v₋,eps₋,p₋)
    tau₊, tau₋    = Formulae.τ(rho₊,v₊,eps₊,p₊), Formulae.τ(rho₋,v₋,eps₋,p₋)
    vv            = absmax(max_v[k₊], max_v[k₋])
    f₊, f₋        = Formulae.flux_D(rho₊,v₊,eps₊,p₊), Formulae.flux_D(rho₋,v₋,eps₋,p₋)
    nflx_D_L[k]   = LLF(nx*f₊, nx*f₋, D₊, D₋, vv)
    f₊, f₋        = Formulae.flux_S(rho₊,v₊,eps₊,p₊), Formulae.flux_S(rho₋,v₋,eps₋,p₋)
    nflx_S_L[k]   = LLF(nx*f₊, nx*f₋, S₊, S₋, vv)
    f₊, f₋        = Formulae.flux_τ(rho₊,v₊,eps₊,p₊), Formulae.flux_τ(rho₋,v₋,eps₋,p₋)
    nflx_tau_L[k] = LLF(nx*f₊, nx*f₋, tau₊, tau₋, vv)
  end
  for k in 1:K-1
    # on right cell interfaces
    nx            = 1
    k₊, k₋        = k, k+1
    v₊, v₋        = vR[k₊], vL[k₋]
    rho₊, rho₋    = rhoR[k₊], rhoL[k₋]
    eps₊, eps₋    = epsR[k₊], epsL[k₋]
    p₊, p₋        = pR[k₊], pL[k₋]
    D₊, D₋        = Formulae.D(rho₊,v₊,eps₊,p₊), Formulae.D(rho₋,v₋,eps₋,p₋)
    S₊, S₋        = Formulae.S(rho₊,v₊,eps₊,p₊), Formulae.S(rho₋,v₋,eps₋,p₋)
    tau₊, tau₋    = Formulae.τ(rho₊,v₊,eps₊,p₊), Formulae.τ(rho₋,v₋,eps₋,p₋)
    vv            = absmax(max_v[k₊], max_v[k₋])
    f₊, f₋        = Formulae.flux_D(rho₊,v₊,eps₊,p₊), Formulae.flux_D(rho₋,v₋,eps₋,p₋)
    nflx_D_R[k]   = LLF(nx*f₊, nx*f₋, D₊, D₋, vv)
    f₊, f₋        = Formulae.flux_S(rho₊,v₊,eps₊,p₊), Formulae.flux_S(rho₋,v₋,eps₋,p₋)
    nflx_S_R[k]   = LLF(nx*f₊, nx*f₋, S₊, S₋, vv)
    f₊, f₋        = Formulae.flux_τ(rho₊,v₊,eps₊,p₊), Formulae.flux_τ(rho₋,v₋,eps₋,p₋)
    nflx_tau_R[k] = LLF(nx*f₊, nx*f₋, tau₊, tau₋, vv)
  end

  # assume periodic bdry conditions
  begin
    k             = 1
    nx            = -1
    k₊, k₋        = periodic_index(k,K), periodic_index(k-1,K)
    v₊, v₋        = vL[k₊], vR[k₋]
    rho₊, rho₋    = rhoL[k₊], rhoR[k₋]
    eps₊, eps₋    = epsL[k₊], epsR[k₋]
    p₊, p₋        = pL[k₊], pR[k₋]
    D₊, D₋        = Formulae.D(rho₊,v₊,eps₊,p₊), Formulae.D(rho₋,v₋,eps₋,p₋)
    S₊, S₋        = Formulae.S(rho₊,v₊,eps₊,p₊), Formulae.S(rho₋,v₋,eps₋,p₋)
    tau₊, tau₋    = Formulae.τ(rho₊,v₊,eps₊,p₊), Formulae.τ(rho₋,v₋,eps₋,p₋)
    vv            = absmax(max_v[k₊], max_v[k₋])
    f₊, f₋        = Formulae.flux_D(rho₊,v₊,eps₊,p₊), Formulae.flux_D(rho₋,v₋,eps₋,p₋)
    nflx_D_L[k]   = LLF(nx*f₊, nx*f₋, D₊, D₋, vv)
    f₊, f₋        = Formulae.flux_S(rho₊,v₊,eps₊,p₊), Formulae.flux_S(rho₋,v₋,eps₋,p₋)
    nflx_S_L[k]   = LLF(nx*f₊, nx*f₋, S₊, S₋, vv)
    f₊, f₋        = Formulae.flux_τ(rho₊,v₊,eps₊,p₊), Formulae.flux_τ(rho₋,v₋,eps₋,p₋)
    nflx_tau_L[k] = LLF(nx*f₊, nx*f₋, tau₊, tau₋, vv)
  end
  begin
    k             = K
    nx            = 1
    k₊, k₋        = periodic_index(k,K), periodic_index(k+1,K)
    v₊, v₋        = vR[k₊], vL[k₋]
    rho₊, rho₋    = rhoR[k₊], rhoL[k₋]
    eps₊, eps₋    = epsR[k₊], epsL[k₋]
    p₊, p₋        = pR[k₊], pL[k₋]
    D₊, D₋        = Formulae.D(rho₊,v₊,eps₊,p₊), Formulae.D(rho₋,v₋,eps₋,p₋)
    S₊, S₋        = Formulae.S(rho₊,v₊,eps₊,p₊), Formulae.S(rho₋,v₋,eps₋,p₋)
    tau₊, tau₋    = Formulae.τ(rho₊,v₊,eps₊,p₊), Formulae.τ(rho₋,v₋,eps₋,p₋)
    vv            = absmax(max_v[k₊], max_v[k₋])
    f₊, f₋        = Formulae.flux_D(rho₊,v₊,eps₊,p₊), Formulae.flux_D(rho₋,v₋,eps₋,p₋)
    nflx_D_R[k]   = LLF(nx*f₊, nx*f₋, D₊, D₋, vv)
    f₊, f₋        = Formulae.flux_S(rho₊,v₊,eps₊,p₊), Formulae.flux_S(rho₋,v₋,eps₋,p₋)
    nflx_S_R[k]   = LLF(nx*f₊, nx*f₋, S₊, S₋, vv)
    f₊, f₋        = Formulae.flux_τ(rho₊,v₊,eps₊,p₊), Formulae.flux_τ(rho₋,v₋,eps₋,p₋)
    nflx_tau_R[k] = LLF(nx*f₊, nx*f₋, tau₊, tau₋, vv)
  end

  @turbo for k in 1:K
    # we have a + in the braces, because we use the nx with fluxes above
    rhs_D[k]   = -(nflx_D_R[k]   + nflx_D_L[k]  )/dl
    rhs_S[k]   = -(nflx_S_R[k]   + nflx_S_L[k]  )/dl
    rhs_tau[k] = -(nflx_tau_R[k] + nflx_tau_L[k])/dl
  end

  if !periodic
    # for shock tube tests we set time derivatives to zero at bdrys
    rhs_D[1] = rhs_D[end] = 0.0
    rhs_S[1] = rhs_S[end] = 0.0
    rhs_tau[1] = rhs_tau[end] = 0.0
  end

  return
end


@inline function limit_slope(a, b, dx, prms)
  if prms.method === :none
    return 0.0
  elseif prms.method === :minmod
    return TCI.minmod(a, b)
  elseif prms.method === :muscl
    return TCI.minmod((a+b)/2, 2*a, 2*b)
  elseif prms.method === :superbee
    return TCI.minmod(TCI.maxmod(a,b),TCI.minmod(2*a,2*b))
  elseif prms.method === :van_albada
    c = dx^3
    return TCI.minmod( ((a^2+c^2)*b+(b^2+c^2)*a)/(a^2+b^2+c^2), 2*a, 2*b )
  elseif prms.method === :van_leer
    return TCI.minmod(2*a*b/(a+b), 2*a, 2*b)
  elseif prms.method === :tvb
    return TCI.minmod_M(dx, prms.tvb_M, (a+b)/2, 2*a, 2*b)
  else
    TODO(prms.method)
  end
end


#######################################################################
#                               DG rhs                                #
#######################################################################


function rhs!(env, P::Project, hrsc::Nothing, mesh::Mesh1d)

  @unpack cache, mesh                = env
  @unpack equation                   = P
  @unpack D, S, tau                  = get_dynamic_variables(cache)
  @unpack flx_D, flx_S, flx_tau,
          max_v, rho, eps, v, p      = get_static_variables(cache)
  @unpack rhs_D, rhs_S, rhs_tau      = get_rhs_variables(cache)
  @unpack nflx_D, nflx_S, nflx_tau,
          bdry_D, bdry_S, bdry_tau,
          bdry_max_v, bdry_v,
          bdry_rho, bdry_eps, bdry_p = get_bdry_variables(cache)

  broadcast_volume!(cons2prim_kastaun, equation, mesh)
  broadcast_volume!(maxspeed, equation, mesh)

  dg1d.interpolate_face_data!(mesh, D, bdry_D)
  dg1d.interpolate_face_data!(mesh, S, bdry_S)
  dg1d.interpolate_face_data!(mesh, tau, bdry_tau)
  dg1d.interpolate_face_data!(mesh, max_v, bdry_max_v)
  dg1d.interpolate_face_data!(mesh, rho, bdry_rho)
  dg1d.interpolate_face_data!(mesh, v, bdry_v)
  dg1d.interpolate_face_data!(mesh, eps, bdry_eps)
  dg1d.interpolate_face_data!(mesh, p, bdry_p)

  broadcast_volume!(flux, equation, mesh)
  broadcast_faces!(llf, equation, mesh)
  broadcast_bdry!(bdryllf, equation, mesh)

  compute_rhs_weak_form!(rhs_D,   flx_D,    nflx_D,   mesh)
  compute_rhs_weak_form!(rhs_S,   flx_S,    nflx_S,   mesh)
  compute_rhs_weak_form!(rhs_tau, flx_tau,  nflx_tau, mesh)

  return
end


function rhs!(env, P::Project, hrsc::HRSC.AbstractReconstruction, ::Mesh1d)

  @unpack cache, mesh                = env
  @unpack equation                   = P
  @unpack D, S, tau                  = get_dynamic_variables(cache)
  @unpack flx_D, flx_S, flx_tau,
          max_v, rho, eps, v, p      = get_static_variables(cache)
  @unpack rhs_D, rhs_S, rhs_tau      = get_rhs_variables(cache)
  @unpack nflx_D, nflx_S, nflx_tau,
          bdry_D, bdry_S, bdry_tau,
          bdry_max_v, bdry_v,
          bdry_rho, bdry_eps, bdry_p = get_bdry_variables(cache)

  HRSC.reconstruct!(D,    flag, hrsc, isperiodic=isperiodic(mesh), m=1e-10, limit_with_boundaries=false)
  HRSC.reconstruct!(S,    flag, hrsc, isperiodic=isperiodic(mesh), limit_with_boundaries=false)
  HRSC.reconstruct!(tau,  flag, hrsc, isperiodic=isperiodic(mesh), m=1e-10, limit_with_boundaries=false)

  broadcast_volume!(cons2prim_kastaun, equation, mesh)
  dg1d.interpolate_face_data!(mesh, D,     bdry_D)
  dg1d.interpolate_face_data!(mesh, S,     bdry_S)
  dg1d.interpolate_face_data!(mesh, tau,   bdry_tau)
  dg1d.interpolate_face_data!(mesh, max_v, bdry_max_v)
  dg1d.interpolate_face_data!(mesh, rho, bdry_rho)
  dg1d.interpolate_face_data!(mesh, v, bdry_v)
  dg1d.interpolate_face_data!(mesh, eps, bdry_eps)
  dg1d.interpolate_face_data!(mesh, p, bdry_p)
  broadcast_volume!(flux, equation, mesh)
  broadcast_faces!(llf, equation, mesh)

  compute_rhs_weak_form!(rhs_D,   flx_D,    nflx_D,   mesh)
  compute_rhs_weak_form!(rhs_S,   flx_S,    nflx_S,   mesh)
  compute_rhs_weak_form!(rhs_tau, flx_tau,  nflx_tau, mesh)

  return
end


function fd_diff_1d!(_df, _f, _x, mesh)
  @unpack Npts = mesh.element
  for (df, f, x) in zip(eachcell(mesh,_df),eachcell(mesh,_f),eachcell(mesh,_x))
    for n in 2:Npts-1
      fl,fm,fr = f[n-1],f[n],f[n+1]
      xl,xm,xr = x[n-1],x[n],x[n+1]
      dll = (xm-xr)/((xl-xm)*(xl-xr))
      dlm = (2*xm-xl-xr)/((xm-xl)*(xm-xr))
      dlr = (xm-xl)/((xr-xl)*(xr-xm))
      df[n] = fl*dll + fm*dlm + fr*dlr
    end
    fl,fm,fr = f[1],f[2],f[3]
    xl,xm,xr = x[1],x[2],x[3]
    dll = (2*xl-xr-xm)/((xl-xm)*(xl-xr))
    dlm = (xl-xr)/((xm-xl)*(xm-xr))
    dlr = (xl-xm)/((xr-xl)*(xr-xm))
    df[1] = fl*dll + fm*dlm + fr*dlr
    fl,fm,fr = f[Npts-2],f[Npts-1],f[Npts]
    xl,xm,xr = x[Npts-2],x[Npts-1],x[Npts]
    dll = (xr-xm)/((xl-xm)*(xl-xr))
    dlm = (xr-xl)/((xm-xl)*(xm-xr))
    dlr = (2*xr-xl-xm)/((xr-xl)*(xr-xm))
    df[Npts] = fl*dll + fm*dlm + fr*dlr
  end
end


function rhs!(env, P::Project, hrsc::HRSC.AbstractArtificialViscosity, ::Mesh1d)

  @unpack cache, mesh = env
  @unpack equation, prms = P
  reg = prms.av_regularization

  if reg === :mono
    rhs_mono!(cache, mesh, equation, P)
  elseif reg === :mono_approx_covariant
    rhs_mono_approx_covariant!(cache, mesh, equation, P)
  else
    TODO("artificial viscosity with av_regularization = $reg")
  end

end


function rhs_mono!(cache, mesh, equation, P)

  @unpack D, S, tau             = get_dynamic_variables(cache)
  @unpack flx_D, flx_S, flx_tau,
          ldg_D, ldg_S, ldg_tau,
          smoothed_mu, max_v,
          rho, v, eps, p        = get_static_variables(cache)
  @unpack rhs_D, rhs_S, rhs_tau = get_rhs_variables(cache)
  @unpack nflx_D, nflx_S, nflx_tau,
          bdry_D, bdry_S, bdry_tau,
          bdry_rho, bdry_v, bdry_eps,
          bdry_p, bdry_max_v,
          bdry_ldg_D, bdry_ldg_S,
          bdry_ldg_tau,
          bdry_smoothed_mu      = get_bdry_variables(cache)
  @unpack x = get_static_variables(cache)

  ## solve auxiliary equation: q + ∂x u = 0
  broadcast_faces!(ldg_nflux, equation, mesh)
  broadcast_bdry!(ldg_bdryllf, equation, mesh)
  # broadcast_boundaryconditions!(mean_flux, ldg_bdryconds, cache, mesh, 0.0)
  compute_rhs_weak_form!(ldg_D,   D,   nflx_D,   mesh)
  compute_rhs_weak_form!(ldg_S,   S,   nflx_S,   mesh)
  compute_rhs_weak_form!(ldg_tau, tau, nflx_tau, mesh)
  # fd_diff_1d!(ldg_D,   D,   x, mesh)
  # fd_diff_1d!(ldg_S,   S,   x, mesh)
  # fd_diff_1d!(ldg_tau, tau, x, mesh)


  broadcast_volume!(cons2prim_kastaun, equation, mesh)
  broadcast_volume!(maxspeed, equation, mesh)

  dg1d.interpolate_face_data!(mesh, D, bdry_D)
  dg1d.interpolate_face_data!(mesh, S, bdry_S)
  dg1d.interpolate_face_data!(mesh, tau, bdry_tau)
  dg1d.interpolate_face_data!(mesh, max_v, bdry_max_v)
  dg1d.interpolate_face_data!(mesh, rho, bdry_rho)
  dg1d.interpolate_face_data!(mesh, v, bdry_v)
  dg1d.interpolate_face_data!(mesh, eps, bdry_eps)
  dg1d.interpolate_face_data!(mesh, p, bdry_p)

  broadcast_volume!(flux, equation, mesh)
  broadcast_volume!(av_flux_mono, equation, mesh)

  dg1d.interpolate_face_data!(mesh, ldg_D, bdry_ldg_D)
  dg1d.interpolate_face_data!(mesh, ldg_S, bdry_ldg_S)
  dg1d.interpolate_face_data!(mesh, ldg_tau, bdry_ldg_tau)
  dg1d.interpolate_face_data!(mesh, smoothed_mu, bdry_smoothed_mu)

  # numerical fluxes
  broadcast_faces!(llf, equation, mesh)
  broadcast_faces!(av_nflux_mono, equation, mesh)
  broadcast_bdry!(bdryllf, equation, mesh)

  compute_rhs_weak_form!(rhs_D,   flx_D,   nflx_D,   mesh)
  compute_rhs_weak_form!(rhs_S,   flx_S,   nflx_S,   mesh)
  compute_rhs_weak_form!(rhs_tau, flx_tau, nflx_tau, mesh)

  return
end


function rhs_mono_approx_covariant!(cache, mesh, equation, P)

  @unpack D, S, tau             = get_dynamic_variables(cache)
  @unpack flx_D, flx_S, flx_tau,
          ldg_D, ldg_S, ldg_tau,
          smoothed_mu, max_v,
          rho, v, eps, p        = get_static_variables(cache)
  @unpack rhs_D, rhs_S, rhs_tau = get_rhs_variables(cache)
  @unpack nflx_D, nflx_S, nflx_tau,
          bdry_D, bdry_S, bdry_tau,
          bdry_rho, bdry_v, bdry_eps,
          bdry_p, bdry_max_v,
          bdry_ldg_D, bdry_ldg_S, bdry_ldg_tau,
          bdry_rhs_D, bdry_rhs_S, bdry_rhs_tau,
          bdry_smoothed_mu      = get_bdry_variables(cache)
  @unpack x = get_static_variables(cache)

  # compute rhs for ∂t u + ∂x f(x) = 0

  broadcast_volume!(cons2prim_kastaun, equation, mesh)
  broadcast_volume!(maxspeed, equation, mesh)

  dg1d.interpolate_face_data!(mesh, D,     bdry_D)
  dg1d.interpolate_face_data!(mesh, S,     bdry_S)
  dg1d.interpolate_face_data!(mesh, tau,   bdry_tau)
  dg1d.interpolate_face_data!(mesh, max_v, bdry_max_v)
  dg1d.interpolate_face_data!(mesh, rho,   bdry_rho)
  dg1d.interpolate_face_data!(mesh, v,     bdry_v)
  dg1d.interpolate_face_data!(mesh, eps,   bdry_eps)
  dg1d.interpolate_face_data!(mesh, p,     bdry_p)

  broadcast_volume!(flux, equation, mesh)

  # numerical fluxes
  broadcast_faces!(llf,    equation, mesh)
  broadcast_bdry!(bdryllf, equation, mesh)

  compute_rhs_weak_form!(rhs_D,   flx_D,   nflx_D,   mesh)
  compute_rhs_weak_form!(rhs_S,   flx_S,   nflx_S,   mesh)
  compute_rhs_weak_form!(rhs_tau, flx_tau, nflx_tau, mesh)

  # now compute rhs for ∂t u + ∂x f(x) = ∂x( μ (-∂t+∂x) u )

  ## solve auxiliary equation: q + ∂x u = 0
  broadcast_faces!(ldg_nflux,  equation, mesh)
  broadcast_bdry!(ldg_bdryllf, equation, mesh)
  compute_rhs_weak_form!(ldg_D,   D,   nflx_D,   mesh)
  compute_rhs_weak_form!(ldg_S,   S,   nflx_S,   mesh)
  compute_rhs_weak_form!(ldg_tau, tau, nflx_tau, mesh)
  # fd_diff_1d!(ldg_D,   D,   x, mesh)
  # fd_diff_1d!(ldg_S,   S,   x, mesh)
  # fd_diff_1d!(ldg_tau, tau, x, mesh)

  # need to recompute, because ldg computation overwrites nflx_{D,S,tau}
  # numerical fluxes
  broadcast_faces!(llf,    equation, mesh)
  broadcast_bdry!(bdryllf, equation, mesh)

  broadcast_volume!(av_flux_approx_mono_covariant, equation, mesh)

  dg1d.interpolate_face_data!(mesh, ldg_D,       bdry_ldg_D)
  dg1d.interpolate_face_data!(mesh, ldg_S,       bdry_ldg_S)
  dg1d.interpolate_face_data!(mesh, ldg_tau,     bdry_ldg_tau)
  dg1d.interpolate_face_data!(mesh, rhs_D,       bdry_rhs_D)
  dg1d.interpolate_face_data!(mesh, rhs_S,       bdry_rhs_S)
  dg1d.interpolate_face_data!(mesh, rhs_tau,     bdry_rhs_tau)
  dg1d.interpolate_face_data!(mesh, smoothed_mu, bdry_smoothed_mu)

  broadcast_faces!(av_nflux_approx_mono_covariant, equation, mesh)

  compute_rhs_weak_form!(rhs_D,   flx_D,   nflx_D,   mesh)
  compute_rhs_weak_form!(rhs_S,   flx_S,   nflx_S,   mesh)
  compute_rhs_weak_form!(rhs_tau, flx_tau, nflx_tau, mesh)

  return
end


#######################################################################
#                                 2d                                  #
#######################################################################


function dg1d.rhs!(env, P, ::Nothing, ::Mesh2d)

  TODO()

  @unpack cache, mesh                              = env
  @unpack equation                                 = P
  @unpack D, Sx, Sy, tau                           = get_dynamic_variables(cache)
  @unpack flx_D_x, flx_D_y, flx_Sx_x, flx_Sx_y,
          flx_Sy_x, flx_Sy_y, flx_tau_x, flx_tau_y = get_static_variables(cache)
  @unpack rhs_D, rhs_Sx, rhs_Sy, rhs_tau           = get_rhs_variables(cache)
  @unpack nflx_D, nflx_Sx, nflx_Sy, nflx_tau       = get_bdry_variables(cache)

  broadcast_volume!(cons2prim, equation, mesh)
  broadcast_volume!(flux, equation, mesh)
  broadcast_faces!(llf_D, equation, mesh)
  broadcast_faces!(llf_Sx, equation, mesh)
  broadcast_faces!(llf_Sy, equation, mesh)
  broadcast_faces!(llf_tau, equation, mesh)

  compute_rhs_weak_form!(rhs_D,   flx_D_x,    flx_D_y,   nflx_D,   mesh)
  compute_rhs_weak_form!(rhs_Sx,  flx_Sx_x,   flx_Sx_y,  nflx_Sx,  mesh)
  compute_rhs_weak_form!(rhs_Sy,  flx_Sy_x,   flx_Sy_y,  nflx_Sy,  mesh)
  compute_rhs_weak_form!(rhs_tau, flx_tau_x,  flx_tau_y, nflx_tau, mesh)

  return
end


function dg1d.rhs!(env, P::Project2d, hrsc::HRSC.AbstractArtificialViscosity, mesh::Mesh2d)

  TODO()

  @unpack equation                       = P
  @unpack D, Sx, Sy, tau                 = get_dynamic_variables(mesh.cache)
  @unpack rhs_D, rhs_Sx, rhs_Sy, rhs_tau = get_rhs_variables(mesh.cache)
  @unpack flx_D_x, flx_D_y,
          flx_Sx_x, flx_Sx_y,
          flx_Sy_x, flx_Sy_y,
          flx_tau_x, flx_tau_y,
          ldg_D_x, ldg_D_y,
          ldg_Sx_x, ldg_Sx_y,
          ldg_Sy_x, ldg_Sy_y,
          ldg_tau_x, ldg_tau_y               = get_static_variables(mesh.cache)
  @unpack nflx_D, nflx_Sx, nflx_Sy, nflx_tau,
          nflx_D_x, nflx_D_y,
          nflx_Sx_x, nflx_Sx_y,
          nflx_Sy_x, nflx_Sy_y,
          nflx_tau_x, nflx_tau_y             = get_bdry_variables(mesh.cache)
  @unpack t = get_global_variables(mesh.cache)

  ## TODO Maybe implement a macro with this syntax
  # @weakform(mesh) ∂t ldg_D_x + ∂x D = [ nflx_D_x ]
  # -> compute_rhs_weak_form(ldg_D_x, D, 0, nflx_D_x, mesh)
  # @weakform(mesh) ∂t ldg_D_y + ∂y D = [ nflx_D_y ]
  # @weakform(mesh) ∂t ldg_Sx_x  + ∂x Sx  = [ nflx_Sx_x ]
  # @weakform(mesh) ∂t ldg_Sx_y  + ∂y Sx  = [ nflx_Sx_y ]
  # @weakform(mesh) ∂t ldg_Sy_x  + ∂x Sy  = [ nflx_Sy_x ]
  # @weakform(mesh) ∂t ldg_Sy_y  + ∂y Sy  = [ nflx_Sy_y ]
  # @weakform(mesh) ∂t ldg_tau_x   + ∂x tau   = [ nflx_tau_x ]
  # @weakform(mesh) ∂t ldg_tau_y   + ∂y tau   = [ nflx_tau_y ]

  ## solve auxiliary equation: q + ∂x u = 0
  broadcast_faces!(ldg_nflux, equation, mesh)
  broadcast_bdry!(bdry_ldg_nflux, equation, P.bdrycond, mesh)

  # broadcast_boundaryconditions!(central_flux, ldg_bdryconds, cache, mesh, 0.0)
  compute_rhs_weak_form!(ldg_D_x,   D,   0,   nflx_D_x,   mesh)
  compute_rhs_weak_form!(ldg_D_y,   0,   D,   nflx_D_y,   mesh)
  compute_rhs_weak_form!(ldg_Sx_x,  Sx,  0,   nflx_Sx_x,  mesh)
  compute_rhs_weak_form!(ldg_Sx_y,  0,   Sx,  nflx_Sx_y,  mesh)
  compute_rhs_weak_form!(ldg_Sy_x,  Sy,  0,   nflx_Sy_x,  mesh)
  compute_rhs_weak_form!(ldg_Sy_y,  0,   Sy,  nflx_Sy_y,  mesh)
  compute_rhs_weak_form!(ldg_tau_x, tau, 0,   nflx_tau_x, mesh)
  compute_rhs_weak_form!(ldg_tau_y, 0,   tau, nflx_tau_y, mesh)

  ## compute rhs of regularized equation: ∂t u + ∂x f + ∂x μ q = 0
  broadcast_volume!(cons2prim, equation, mesh)
  broadcast_volume!(flux, equation, mesh)
  broadcast_faces!(llf, equation, mesh)
  broadcast_bdry!(bdry_llf, equation, P.bdrycond, mesh)

  broadcast_volume!(av_flux_mono, equation, mesh)
  broadcast_faces!(av_nflux_mono, equation, mesh)
  broadcast_bdry!(bdry_av_nflux_mono, equation, P.bdrycond, mesh)

  compute_rhs_weak_form!(rhs_D,   flx_D_x,   flx_D_y,   nflx_D,   mesh)
  compute_rhs_weak_form!(rhs_Sx,  flx_Sx_x,  flx_Sx_y,  nflx_Sx,  mesh)
  compute_rhs_weak_form!(rhs_Sy,  flx_Sy_x,  flx_Sy_y,  nflx_Sy,  mesh)
  compute_rhs_weak_form!(rhs_tau, flx_tau_x, flx_tau_y, nflx_tau, mesh)

  return
end
